import '../route_mappers.dart';
import '../router.dart';
import 'models.dart';
import 'utils.dart';
import 'interfaces.dart';
import 'url_utils.dart';

class Matcher {
  /// 路由与 Uri 是否匹配，如果不匹配就返回 Null
  static Map match(
      Route route, Uri uri, EqualChecker checker) {
    if (checkSchemeAndHostEqual(route, uri, checker) &&
        uri.pathSegments.length == route.path.length) {
      for (int i = 0; i < route.path.length; i++) {
        Segment segment = route.path[i];
        if (segment.isParam) {
          if (!checkSegmentValid(uri.pathSegments[i], segment)) {
            return null;
          }
        } else if (segment.key != uri.pathSegments[i]) {
          return null;
        }
      }
      return getMatchedParams(route, uri);
    }
    return null;
  }

  /**
   * 判断 scheme 和 host 是否 match，比如说我们可能认为 zhihu.dev == zhihu.com,http == https
   */
  static bool checkSchemeAndHostEqual(
      Route route, Uri uri, EqualChecker checker) {
    String templateScheme = getSchemeOrDefault(route);
    String templateHost = route.host;
    return checkSchemeAndHostEquals(
        uri.scheme, templateScheme, uri.host, templateHost, checker);
  }

  /// 判断 scheme 和 host 是否 match，比如说我们可能认为 zhihu.dev == zhihu.com,http == https
  static bool checkSchemeAndHostEquals(String scheme, String templateScheme,
      String host, String templateHost, EqualChecker checker) {
    if (checker == null ||
        !checker.intercept(scheme, templateScheme, host, templateHost)) {
      return scheme == templateScheme && host == templateHost;
    } else {
      return checker.checkEquals(scheme, templateScheme, host, templateHost);
    }
  }

  /// 返回一个 bundle ，将参数都塞进去
  static Map getMatchedParams(Route route, Uri uri) {
    //走到此处说明 path 已经 match
    Map bundle = {};
    for (int i = 0; i < route.path.length; i++) {
      Segment segment = route.path[i];
      if (segment.isParam) {
        String value = uri.pathSegments[i];
        putExtra(bundle, segment.key, value, segment.type);
      }
    }

    var queryParameterNames = uri.queryParameters.keys;
    for (var entry in route.queries.entries) {
      Query query = entry.value;
      String queryKey = query.queryKey;
      String bundleKey = query.bundleKey;
      if (queryParameterNames.contains(queryKey)) {
        String value = uri.queryParameters[queryKey];
        if (checkValid(value, query)) {
          putExtra(bundle, bundleKey, value, query.type);
        } else {
          return null;
        }
      } else if (!isEmpty(query.value)) {
        //如果 template 中已经写好一个 k-v 了，那么就直接把这个 k-v 放到 bundle 中
        if (checkValid(query.value, query)) {
          putExtra(bundle, bundleKey, query.value, query.type);
        }
      } else if (query.isParam && !query.nullable) {
        //如果 template 中的 query 为参数且无默认值又不可为空，而传入的 uri 又没有这个 query，就不匹配
        return null;
      }
    }
    putQueriesIntoBundle(uri, queryParameterNames, bundle, route);
    if (route.redirect != null) {
      bundle[Router.KEY_ROUTER_RAW_URL] = redirectRoute(route.redirect, bundle);
      bundle[Router.KEY_ROUTER_REDIRECT_FROM] = uri.toString();
    } else {
      bundle[Router.KEY_ROUTER_RAW_URL] = uri.toString();
    }
    return bundle;
  }

  /**
   * 将 url 中的 query 的（template中未提及的）值传入 bundle
   */
  static void putQueriesIntoBundle(Uri uri, Iterable<String> queryNames,
      Map bundle, Route route) {
    // 虽然 queryNames 可以通过 uri.getQueryParameterNames()，但是直接传入的效率更高
    for (var key in queryNames) {
      //如果  bundle 中不存在并且 template 中不存在，就加入
      if (!bundle.containsKey(key) && !route.queries.containsKey(key)) {
        putExtra(bundle, key, uri.queryParameters[key], "string");
      }
    }
  }

  /**
   * 重定向到另一个 Url
   */
  static String redirectRoute(Route redirect, Map bundle) {
    // 提取出被转向的 Url 中的所有参数
    Map redirectParams = rebuildBundle(redirect, bundle);

    // 获取完整的 Url
    String redirectRawUrl = UriUtils.fixUrlScheme(
        RouterMappers.DEFAULT_SCHEME, redirect.template);

    // 将 url 中的 {xx} 替换成真实值
    return UriUtils.replaceParamInUrl(redirectParams, redirectRawUrl);
  }

  static Map rebuildBundle(Route redirect,
      Map bundle) {
    Map redirectParamMap = {};
    for (var segment in redirect.path) {
      if (segment.isParam) {
        String value = getPreciseString(bundle, segment.key, segment.value);
        // 由于 RouterProcessor 进行了前期检查，所以此时 value 必然不为空，否则 route 本身就匹配不过去，type 也必然是 match 的
        putExtra(bundle, segment.key, value, segment.type);
        redirectParamMap[segment.key] = value;
      }
    }

    for (var entry in redirect.queries.entries) {
      Query query = entry.value;
      if (query.isParam) {
        String value = getPreciseString(bundle, query.bundleKey, query.value);
        // type 应该是 match 的
        // 如果 query.nullable=false，那么 value 必然不为空，如果 query 是 nullable，那么 value 就有可能是空的
        // value != null == !query.nullable || (query.nullable && value!=null)
        if (value != null) {
          putExtra(bundle, query.bundleKey, value, query.type);
          redirectParamMap[query.bundleKey] = value;
        }
      } else {
        //如果 template 中已经写好一个 k-v 了，那么就直接把这个 k-v 放到 bundle 中，无论是否是参数化的
        putExtra(bundle, query.bundleKey, query.value, query.type);
      }
    }

    return redirectParamMap;
  }

  static String getPreciseString(Map bundle, String key,
      String fallback) {
    Object value = bundle[key];
    if (value == null) {
      return fallback;
    }
    //然而并不准，不过也没有这个需求
    return value.toString();
  }

  /**
   * 检查传入的值与预先在注解里面定义的是否匹配，包含正则检查和类型检查
   */
  static bool checkSegmentValid(String value, Segment segment) {
    return checkRegValid(value, segment.type, segment.reg);
  }

  /// 检查传入的值与预先在注解里面定义的是否匹配，包含正则检查和类型检查
  static bool checkValid(String value, Query query) {
    // 如果 query 为可空类型并且值就是空的，那么是合法的
    // 否则就要走 checkValid 检查类型
    return query.nullable && isEmpty(value) ||
        checkRegValid(value, query.type, query.reg);
  }

  /// 检查一个是否与其正则和类型匹配
  static bool checkRegValid(String value, String type, String reg) {
    return (isEmpty(reg) || value.matches(reg)) && checkType(value, type);
  }

  static void putExtra(Map bundle, String key, String value,
      String type) {
    // 编译期已经确保了 type 不为空
    switch (type) {
      case "int":
        bundle[key] = int.parse(value);
        break;
      case "bool":
        bundle[key] = value == "true";
        break;
      case "double":
        bundle[key] = double.parse(value);
        break;
      default:
        bundle[key] = value;
        break;
    }
  }

  static String getSchemeOrDefault(Route route) {
    if (route.scheme == null) {
      return RouterMappers.DEFAULT_SCHEME;
    }
    return route.scheme;
  }
}
