import 'dart:collection';

class Route {
  /// 路由 url 模板，如果是出于节约考虑，这个字段可以没有
  final String template;

  // 路由的 scheme
  final String scheme;

  // 路由的 host
  final String host;

  // 路由的 path，即 url host 后面 query 前面的东西
  final List<Segment> path;

  // 路由的 query
  final Map<String, Query> queries;

  Route redirect;

  /// 由 Processor 保证不会出现 null 值（redirect除外）
  Route(this.template, this.scheme, this.host, this.path, this.queries,
      this.redirect);

  @override
  String toString() {
    return """Route{
    rawUrl: $template, 
    scheme: $scheme, 
    host: $host, 
    path: $path, 
    queries: $queries, 
    redirect: $redirect
  }""";
  }
}


class Segment {
  /// 写在 Router 注解中的 path 对应的值
  ///
  /// 比如 zhihu://question/{question_id}/answer/{answer_id} 有三个 segment ，它们的 raw 分别是：
  /// 1. "{question_id}"
  /// 2. "answer"
  /// 3. "{answer_id}"
  ///
  /// 出于节约考虑，可以没有
  final String raw;

  /// 如果是 isParam 为 true，那么它是放到 bundle 里面的 key 的名字
  ///
  /// 比如 zhihu://question/{question_id}/answer/{answer_id} 有三个 segment ，它们的 key 分别是：
  /// 1. "question_id"
  /// 2. "answer"
  /// 3. "answer_id"
  final String key;

  /// 默认值或者真实值
  ///
  /// 如果 isParam 为 false 那么 value == key。
  /// 如果 isParam 为 true 那么 value 就是最后从真实 url 中匹配出来的值。
  ///
  /// 比如 zhihu://question/{question_id}/answer/{answer_id} 来匹配 zhihu://question/233/answer/111
  ///
  /// 三个 Segment 的 value 分别为 233、answer、111
  final String value;

  /// 参数类型，是以下值之一：int|bool|double|string
  ///
  /// 绝对不可能为空
  final String type;

  /// 表示这个 Segment 是否是一个参数，用于捕获捕获 url 中的值
  ///
  /// 比如 zhihu://question/{question_id}/answer/{answer_id} 来匹配 zhihu://question/233
  ///
  /// 那么 question_id 和 answer_id 的 isParam 为 true，而 answer 的为 false
  final bool isParam;

  /// 这个 Segment 对应的正则表达式
  final String reg;

  const Segment(this.raw, this.key, this.value, this.type, this.isParam, this.reg);

  @override
  String toString() {
    return """Segment{
      raw: $raw, 
      key: $key, 
      value: $value, 
      type: $type, 
      isParam: $isParam, 
      reg: $reg
    }""";
  }
}


class Query {
  // 从路由模板中提取出来的单个 query 部分，比如 queryKey={bundleKey?:string=v}
  // 出于节约考虑，这个字段可以没有
  final String raw;

  /// 写在 url 里面的 key，对应 queryKey={bundleKey?:string=v} 中的 queryKey
  /// 如果不存在 queryKey，那 queryKey == bundleKey
  final String queryKey;

  /// 写在 bundle 里面的 bundleKey,，对应 queryKey={bundleKey?:string=v} 中的 bundleKey
  final String bundleKey;

  /// 该 query 的值，如果是参数化的，那么它应该与默认值相等，即在 queryKey={key?:string=v} 的情况下，它的值为 v
  /// 如果是非参数化的，那么应该是 query 真正的 value，即在 q=vv 的情况下，它的值为 vv
  final String value;

  /// 参数类型，绝对不可能为空，默认为 string
  final String type;

  /// 该 query 是否是参数化的
  final bool isParam;

  /// 该 query 是否可以为空
  final bool nullable;

  /// 该 query 的 value 所必须符合的正则表达式
  final String reg;

  const Query(this.raw, this.queryKey, this.bundleKey, this.value, this.type,
      this.isParam, this.nullable, this.reg);

  @override
  String toString() {
    return """Query{
        raw: $raw, 
        queryKey: $queryKey, 
        bundleKey: $bundleKey,
        value: $value,
        type: $type,
        isParam: $isParam,
        nullable: $nullable,
        reg: $reg
      }""";
  }
}
