library router;

import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';

import 'resolver/interfaces.dart';
import 'resolver/match_result.dart';
import 'resolver/matcher.dart';
import 'resolver/url_utils.dart';
import 'resolver/utils.dart';
import 'route_mappers.dart';

class Router {
  /// bundle 中的 key， 原始 url，如果有跳转，则是跳转后的 url
  static final KEY_ROUTER_RAW_URL = "key_router_raw_url";

  /// bundle 中的 key， 如果有跳转，它才是真正的原始 url
  static final KEY_ROUTER_REDIRECT_FROM = "key_router_redirect_from";

  static final KEY_PREVIOUS_ARGUMENTS = "KEY_PREVIOUS_ARGUMENTS";

  /// 在本类的静态代码块中初始化，不存在并发问题
  static final List<Mapper> sMapperList = [];

  static final List<Filter> mFilters = [];

  static EqualChecker mChecker;

  /// 添加一个全局的 Filter，当前 Router 的所有路由匹配都会用到它，任意一个 Filter 返回了 False，就算不匹配了
  static void addFilter(Filter filter) {
    if (filter == null) {
      return;
    }
    mFilters.add(filter);
  }

  /// 设置一个 host 和 scheme 检查器
  static void setChecker(EqualChecker checker) {
    mChecker = checker;
  }

  /// 将 url 解析成 MatchResult，如果不匹配，返回空
  static MatchResult resolve(String url) {
    if (isEmpty(url)) {
      return null;
    }
    return resolveUri(Uri.parse(url));
  }

  /// 将 uri 解析成 MatchResult，如果不匹配，返回空
  static MatchResult resolveUri(Uri uri) {
    if (uri == null) {
      return null;
    }
    Map bundle;

    if (RouterMappers.DEFAULT_SCHEME != null && isEmpty(uri.scheme)) {
      // 如果 scheme 是 null，就给它一个默认的 scheme
      uri = UriUtils.fixUriScheme(RouterMappers.DEFAULT_SCHEME, uri);
    }

    for (var m in sMapperList) {
      bundle = Matcher.match(m.route, uri, mChecker);
      if (bundle != null) {
        MatchResult result =
            MatchResult(uri.toString(), bundle, m.target, m.module);
        if (filter(result)) {
          return result;
        }
      }
    }
    return null;
  }

  /// 只要 filter 中有一个返回了 false ，则说明不符合
  ///
  /// 因为添加的时候就已经判空了，所以这里不再判空
  static bool filter(MatchResult result) {
    for (var filter in mFilters) {
      if (!filter.test(result)) {
        return false;
      }
    }
    return true;
  }

  /// 解析 settings 出路由
  static RouteFactory getOnGenerateRoute() {
    return (RouteSettings settings) => _onGenerateRoute(settings);
  }

  /// 解析 settings 出路由
  static Route<dynamic> _onGenerateRoute(RouteSettings rs) {
    if (notEmpty(rs.name)) {
      MatchResult result = resolve(rs.name);
      if (result != null) {
        if (rs.arguments != null) {
          // 将 RouteSettings.arguments 的数据合并
          if (rs.arguments is Map) {
            // 如果 arguments 是 Map，那就添加到 result.arguments 中去
            result.arguments.addAll(rs.arguments);
          } else {
            // 否则添加到另一个新 key 中去：KEY_PREVIOUS_ARGUMENTS
            result.arguments[KEY_PREVIOUS_ARGUMENTS] = rs.arguments;
          }
        }
        RouteSettings settings = RouteSettings(
            name: rs.name,
            isInitialRoute: rs.isInitialRoute,
            arguments: result.arguments);
        WidgetBuilder builder = (c) => result.builder(result.arguments);
        return MaterialPageRoute(builder: builder, settings: settings);
      }
    }
    return null;
  }
}
