import 'dart:collection';

import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:zrouter/resolver/utils.dart';
import 'package:zrouter/src/compiler/route_element.dart';

import '../../annotations.dart';
import '../../resolver/models.dart';
import 'extensions.dart';

String locateElement(Element element, BuildStep buildStep) {
  final path = buildStep.inputId.path;
  final package = buildStep.inputId.package;
  final import = "package:$package/${path.replaceFirst('lib/', '')}";
  return import + " : " + element.displayName;
}

class Logger {
  RouteElement element;
  StringBuffer buffer;

  Logger(this.element, this.buffer);

  void error(String msg) {
    buffer.writeln("error: $msg");
  }

  void debug(String msg) {}
}

class ClassDefinition {
  String importString;
  String className;
  String importShort;
}

class InnerMapper {
  final String template;
  final Route route;
  final RouteElement target;
  final int priority;
  final String module;

  InnerMapper(
      this.target, this.template, this.route, this.priority, this.module);

  String getBuilder(String name) {
    String id = target.importString + ":" + target.displayName;
    // 此处不对哟
    String className = target.importShort + "." + target.displayName;
    // 函数的参数 key
    Set<String> names;
    if (target.element is FunctionElement) {
      // 如果注解的是一个函数，那么直接检查即可
      FunctionElement e = target.element;
      names = getParamNames(e.type, id);
      if (names == null) {
        throw Exception("$id 注解的函数与路由参数不匹配");
      }

      if (target.displayName.contains("Cons")) {
        print("===----=++");
        print(e.displayName);
        print(e.displayName);
      }
    } else if (target.element is ClassElement) {
      ClassElement e = target.element;
      for (var value in e.constructors) {
        names = getParamNames(value.type, id);
        if (names != null) {
          if (notEmpty(value.name)) {
            className = target.importShort +
                "." +
                target.displayName +
                "." +
                value.name;
          }
          break;
        }
      }
      if (names == null) {
        throw Exception("$id 注解的函数与路由参数不匹配");
      }
    }

    StringBuffer sb = new StringBuffer();
    for (var name in names) {
      sb..write("$name: m['$name'],");
    }
    return "$name = (m) => ${className}(${sb.toString()});";
  }

  /// 路由中声明的类型与方法参数是否一致
  bool typeMatch(String routeType, DartType type) {
    // 暂时忽略 dynamic 类型，有需要了再说
    return routeType == "string" && type.isDartCoreString ||
        routeType == "int" && type.isDartCoreInt ||
        routeType == "bool" && type.isDartCoreBool ||
        routeType == "double" && type.isDartCoreDouble;
  }

  /// 路由参数是否与某一 function 相匹配，如果匹配就返回对照的 param name
  /// 1. 必须都是命名参数
  /// 2. 命名参数中的参数名，在 route 中必须存在，除非是可选参数
  Set<String> getParamNames(FunctionType type, String id) {
    // 要么命名参数是匹配的，要么参数是 Map
    // 如果命名参数为空，那么随意了
    Set<String> names = {};
    for (var v in type.parameters) {
      if (!v.isNamed) {
        print("$id 参数必须都是 namedParameters");
        return null;
      }

      // 参数是否是不可空的
      bool isRequired = v.isReallyRequiredNamed();
      // 表示 v 是否与路由中的参数相匹配
      bool matches = false;

      // 检查 path 中是否有与之匹配的，path 中所有参数都是不可空的
      for (var value in route.path.where((i) => i.isParam && i.key == v.name)) {
        if (typeMatch(value.type, v.type)) {
          matches = true;
        } else {
          // 如果不相等，直接挂
          print("$id 参数 ${v.name} 与路由类型不匹配");
          return null;
        }
      }

      // 如果 path 中未找到，再找 query 中是否有匹配的
      if (!matches) {
        for (var value in route.queries.values
            .where((i) => i.isParam && i.bundleKey == v.name)) {
          if (typeMatch(value.type, v.type)) {
            matches = !isRequired || !value.nullable;
            if (!matches) {
              print("$id 参数 ${v.name} 与路由可空性不匹配");
            }
          } else {
            print("$id 参数 ${v.name} 与路由可空性不匹配");
            return null;
          }
        }
      }
      if (!matches && isRequired) {
        // 此时如果 !matches 说明啥，说明没有找到此对应的，连名字一样的都没有
        // 但是，如果此参数 isRequired 为 true，就势必要返回 false 了
        // 但是如果 isRequired 为 false，说明没找到也没啥，毕竟不是必填的
        print("$id ${v.name} 没找到且不可空");
        return null;
      }
      if (matches) {
        // 只会对找到参数的添加
        names.add(v.name);
      }
    }
    // 上面没有返回 false，这里就必然为 true 了
    return names;
  }

  @override
  String toString() {
    return """
InnerMapper{
  target: $target,
  template: $template, 
  route: $route, 
  priority: $priority, 
  module: $module
}""";
  }
}

/// 将一个注解 [ConstantReader] 还原成 ScreenRoute，此时 ScreenRoute.routes 必然为 set
ScreenRoute fromAnnotation(Logger logger, ConstantReader annotation) {
  // 读取定义的路由模板
  Set<String> templates = {};
  ConstantReader route = annotation.read("route");
  if (route.isString) {
    templates.add(route.stringValue);
  } else {
    var obj = route.objectValue;
    if (obj.type.isDartCoreSet && obj.type.toString() == "Set<String>") {
      for (var value in obj.toSetValue()) {
        templates.add(value.toStringValue());
      }
    } else {
      logger.error(
          "route 只能是 String 或者 Set<String>，当前类型是：${route.objectValue.type
              .toString()}");
    }
  }

  if (templates.isEmpty) {
    logger.error("定义的路由为空");
  }

  if (templates.where((t) => isEmpty(t)).length > 0) {
    logger.error("路由不能使用空字符串");
  }

// 读取定义的路由参数正则
  Map<String, String> paramReg = {};
  ConstantReader paramRegReader = annotation.read("paramReg");
  var map = paramRegReader.mapValue;
  for (var k in map.keys) {
    paramReg[k.toStringValue()] = map[k].toStringValue();
  }

// 读取路由优先级
  int priority = annotation.read("priority").intValue;

  return ScreenRoute(templates, paramReg: paramReg, priority: priority);
}
