import 'dart:collection';

import 'package:zrouter/src/compiler/route_element.dart';

import '../../resolver/models.dart';
import 'compiler_util.dart';
import 'part_creator.dart';
import 'processor.dart';

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

class MapperCreator {
  static bool isEmpty(String str) {
    return str == null || str.isEmpty;
  }

  /// 生成一个 {@link InnerMapper} 对象，用于表示一条路由数据
  static InnerMapper generateMapperEntity(
      Logger logger,
      RouteElement r,
      String uri,
      Element cls,
      Map<String, String> regs,
      int priority,
      String module) {
    if (isEmpty(uri)) {
      logger.error("路由模板不能为空 ");
      return null;
    }

    // 路由的 url
    String mainUri;
    // 路由 url 可能会有重定向，这是要重定向到的 Uri
    String redirectUri;
    //替换所有的空格，并以 => 分开，重定向用
    List<String> urls = uri.replaceAll("\\s+", "").split("=>");
    mainUri = urls[0];

    Route redirectedRoute = null;
    Route route;
    if (urls.length == 2) {
    // 如果声明了重定向 =>
      redirectUri = urls[1];
      redirectedRoute = generateRouteEntity(logger, regs, redirectUri, null);
      if (redirectedRoute == null) {
        logger.error("错误的重定向路由模板 :  " + uri);
      }
    }

    route = generateRouteEntity(logger, regs, mainUri, redirectedRoute);

    if (route == null) {
      logger.error("错误的路由模板  :  " + uri);
      return null;
    }

    if (route.redirect != null && !match(route, route.redirect, logger)) {
      logger.error("Route 和重定向的 Route 不匹配  :  $uri");
      return null;
    }

    return new InnerMapper(r, uri, route, priority, module);
  }

  /**
   * 生成一个 Route 对象
   */
  static Route generateRouteEntity(Logger logger,
      Map<String, String> paramRegs, String uri, Route redirect) {
    if (isEmpty(uri)) {
      logger.error("bar uri : empty");
      return null;
    }
    String scheme = null;
    String host;
    String hostPathString;
    String queryString = null;

// scheme
    int i = uri.indexOf("://");
    if (i > 0) {
      scheme = uri.substring(0, uri.indexOf("://"));
    }

// host + path 的开头 index （就是 host 的 index）
    int hostPathStartIndex =
        getSchemeSpecificPartStartIndex(uri, scheme, logger);

// host + path 的结尾 index+1，（就是 path 最后一个字母的 index+1）
    int hostPathEndIndex = uri.indexOf("?");

    bool hasQuery = hostPathEndIndex > 0;
    if (!hasQuery) {
      hostPathEndIndex = uri.length;
    }

// host + path，但是在路由里面不重要，都一样对待了
    hostPathString = uri
        .substring(hostPathStartIndex, hostPathEndIndex)
        .replaceAll(RegExp("/+"), "/")
        .replaceAll(RegExp(r"(^/+)|(/+$)"), "");

    host = hostPathString.split("/")[0];

    if (host == null || host.length == 0) {
      logger.error("缺少 host: " + uri);
      return null;
    }

    // TODO Host Only

    // 将头尾的斜杠（/）删除
    String pathString = hostPathString
        .substring(host.length)
        .replaceAll(RegExp(r"(^/+)|(/+$)"), "");

    if (hasQuery) {
      queryString = uri.substring(hostPathEndIndex + 1);
    }

// 根据 pathString 生成一串 Segment
    List<Segment> segmentList = [];
    if (!isEmpty(pathString)) {
      List<String> segments = pathString.split("/");
      for (var s in segments) {
// 创建 segment 对象
        Segment segment = PartCreator.createSegment(logger, s, paramRegs);
        if (segment == null) {
          logger.error("错误的 segment : " + s);
          return null;
        }
        segmentList.add(segment);
      }
    }

    // 根据 queryString 生成一串 Query
    LinkedHashMap<String, Query> queryLinkedHashMap = new LinkedHashMap();
    if (!isEmpty(queryString)) {
      List<String> queries = queryString.split("&");
      for (var s in queries) {
        Query query = PartCreator.createQuery(logger, s, paramRegs);
        if (query == null) {
          logger.error("错误的 query : " + s);
          return null;
        }
        queryLinkedHashMap[query.queryKey] = query;
      }
    }

    logger.debug("start generateRouteEntity : " + uri);
    logger.debug("rawUrl: $uri");
    logger.debug("scheme: $scheme");
    logger.debug("host: $host");
    logger.debug("path: $pathString");
    logger.debug("query: $queryString");
    return new Route(
        uri, scheme, host, segmentList, queryLinkedHashMap, redirect);
  }

  /**
   * 一条 route 是否可以正确的转换为另一个 redirect Route
   */
  static bool match(Route route, Route redirect, Logger logger) {
    for (var segment in redirect.path) {
      if (segment.isParam &&
          (segment.value == null || segment.value.length == 0)) {
//如果 redirect 的某 segment 的默认值为空，那么 route 就必须有非空的参数才成
        if (!hasParam(
            route, segment.key, segment.type, segment.reg, false, logger)) {
          return false;
        }
      }
    }

    for (var entry in redirect.queries.entries) {
      Query query = entry.value;
      if (query.isParam && (query.value == null || query.value.length == 0)) {
        // 如果 redirect 的某 query 的默认值为空，那么 route 就必须有非空的参数才成
        // 但是如果是可空的，那就不用管了，只要保证类型相同就行，
        if (!hasParam(route, query.bundleKey, query.type, query.reg,
            query.nullable, logger)) {
          return false;
        }
      }
    }

    return true;
  }

  /**
   * 一条路由是否包含一个限定的参数
   *
   * @param route    要查找的路由
   * @param key      参数的 key
   * @param type     参数的类型
   * @param reg      参数的颪
   * @param nullable 参数是否可空
   */
  static bool hasParam(Route route, String key, String type, String reg,
      bool nullable, Logger logger) {
    if (type == null || type.trim().length == 0) {
      type = "string";
    }
    type = type.trim();
    for (var segment in route.path) {
      if (segment.isParam &&
          segment.key == key &&
          segment.type == type &&
          segment.reg == reg) {
        return true;
      }
    }

    for (var entry in route.queries.entries) {
      Query query = entry.value;
      if (query.isParam &&
          query.bundleKey == key &&
          query.type == type &&
          query.reg == reg) {
// query.nullable 必须为 false 才能保证 redirect 不接收到空值
        if (nullable) {
// 如果 nullable 为 true，说明 redirect 的 query 可以为空跳过，那么只要类型对就可以了
          return true;
        } else {
// 如果 nullable 为 false，说明 redirect 的 query 不可为空，
// 那么 route 对应的 query 不但类型要对上,nullable 也必须为 false
          if (query.nullable) {
            logger.error("~Route 不匹配 RedirectRoute 的 key: $key, type: $type, reg: $reg");
            return false;
          } else {
            return true;
          }
        }
      }
    }
    logger.error("Route 不匹配 RedirectRoute 的 key: $key, type: $type, reg: $reg");
    return false;
  }

  /**
   * host + path 的开头 index，比如 http://zhihu.com/answer 返回的就是 indexOf("zhihu.com/answer")
   */
  static int getSchemeSpecificPartStartIndex(
      String uri, String scheme, Logger logger) {
    int index;
    if (isEmpty(scheme)) {
      logger.debug("缺少 scheme: " + uri);
//允许 ://host 或者 /host 或者 host开头
      if (uri.startsWith("://")) {
        index = 3;
      } else if (uri.startsWith("//")) {
        index = 2;
      } else if (uri.startsWith("/")) {
        index = 1;
      } else {
        index = 0;
      }
    } else {
      index = scheme.length + 3;
    }
    return index;
  }
}
