import 'dart:collection';

import 'package:zrouter/resolver/utils.dart';

import '../../resolver/models.dart';
import 'compiler_util.dart';
import 'processor.dart';

class PartCreator {
  /// dart 不支持组名
  /// Segment 的正则匹配表达式，匹配的是{key?:type=value}
  static String SEGMENT_REG = r"\{" +
      // 这里是 key 的定义，不能是 =:{} , 当然，可以添加更多
      r"([^=:{}]+)" +
      // 这里是类型的定义，可以没有，样式为 【:类型名】
      r"(:(int|bool|double|string))?" +
      // 这里是默认值的定义，可以没有，样式为 【=值】，值不能是=:{}，当然可以更严格
      r"(=([^=:{}]+))?" +
      // 右花括号结尾
      r"\}";
  static Pattern SEGMENT_PATTERN = RegExp(SEGMENT_REG);

  /// Query 的正则匹配表达式：queryKey={key?:type=value} 或者 {key?:type=value}
  static String QUERY_REG =
      // 开头可能为 queryKey，这里是 url 中真正的 query key
      r"(([^=:{}]+)=)?" +
          // 左花括号形状
          r"\{" +
          // 这里是 bundleKey 的定义，key 是指最终存储到 bundle 中的 key
          // 如果前面没有 queryKey，那么 bundleKey 就是 url 中真正的 key
          // 不能是 =:{} , 当然，可以添加更多
          r"([^=:{}]+)" +
          // 这里是类型的定义，可以没有，样式为 【:类型名】
          r"(:(int|bool|double|string))?" +
          // 这里是默认值的定义，可以没有，样式为 【=值】，值不能是=:{}，当然可以更严格
          r"(=([^=:{}]+))?" +
          // 右花括号结尾
          r"\}";
  static Pattern QUERY_PATTERN = RegExp(QUERY_REG);

  /// 生成一个 Segment 对象
  ///
  /// @param raw       Segment 在模板里面的原始声明，比如 zhihu://answer/{host:int} 第一个 Segment 的 raw 是 {host:int}
  /// @param paramRegs 要匹配的正则们
  static Segment createSegment(
      Logger logger, String raw, Map<String, String> paramRegs) {
    // 如果是符合正则，那么就认为是一个参数化的 Segment，否则就一普通的 Segment
    var matcher = allMatchResult(raw, SEGMENT_REG);
    if (matcher != null) {
      String key = matcher.group(1);
      String type = matcher.group(3);
      String value = matcher.group(5);

      String regexp = paramRegs[key];

      if (notEmpty(regexp) && notEmpty(value)) {
        if (!value.matches(regexp)) {
          logger
              .error("Segment 正则不匹配 : 声明的默认值是 " + value + " 但是对应的正则是 " + type);
          return null;
        }
      }
      type = convertType(type);
      if (checkType(value, type)) {
        return new Segment(raw, key, value, type, true, regexp);
      } else {
        logger.error(
            "Segment 类型不匹配 : 声明的默认值是 " + value + " 但是声明的 type 是 " + type);
        return null;
      }
    } else {
      if (isValidRaw(raw)) {
        return new Segment(raw, raw, raw, "string", false, null);
      } else {
        logger.error("字符串 " + raw + " 格式有问题，可能包含特殊字符");
      }
    }
    return null;
  }

  /// 生成一个 Query 对象
  ///
  /// @param raw       Query 在模板里面的原始声明，比如 zhihu://answer?{host} 第一个 Query 的 raw 是 {host}
  /// @param paramRegs 要匹配的正则们
  static Query createQuery(
      Logger logger, String raw, Map<String, String> paramRegs) {
    // 如果是符合正则，那么就认为是一个参数化的 Query，否则就一普通的 Query
    var matcher = allMatchResult(raw, QUERY_REG);
    if (matcher != null) {
      // queryKey 是传入的 url 使用的 key，而 bundleKey 则是真正最后放到 bundle 里面用的 key
      String queryKey = matcher.group(2);
      String bundleKey = matcher.group(3);
      String type = matcher.group(5);
      String value = matcher.group(7);
      bool nullable = bundleKey.endsWith("?");
      if (nullable) {
        bundleKey = bundleKey.substring(0, bundleKey.length - 1);
      }

      String reg = paramRegs[bundleKey];

      if (notEmpty(reg) && notEmpty(value)) {
        if (!value.matches(reg)) {
          logger.error("Query 正则不匹配 : 声明的默认值是 " + value + " 正则是 " + reg);
          return null;
        }
      }

      type = convertType(type);
      if (queryKey == null) {
        queryKey = bundleKey;
      }
      if (checkType(value, type)) {
        return new Query(
            raw, queryKey, bundleKey, value, type, true, nullable, reg);
      } else {
        logger
            .error("Query 类型不匹配 : 声明的默认值是 " + value + " 但是声明的 type 是 " + type);
        return null;
      }
    } else {
      List<String> kv = raw.split("=");
      if (kv.length > 1) {
        if (isValidRaw(kv[0]) && isValidRaw(kv[1])) {
          return new Query(
              raw, kv[0], kv[0], kv[1], "string", false, false, null);
        } else {
          logger.error("字符串 " + raw + " 格式有问题，可能包含特殊字符");
        }
      } else {
        logger.error("bad Query : " + raw);
      }
    }
    return null;
  }

  static bool checkType(String value, String type) {
    if (value == null) {
      return true;
    }
    try {
      switch (type) {
        case "int":
          int.parse(value);
          break;
        case "double":
          double.parse(value);
          break;
        case "bool":
          String v = value.toLowerCase();
          if ("true" != v || "false" != v) {
            throw FormatException("$value is not bool type");
          }
          break;
        default:
          break;
      }
    } on FormatException {
      return false;
    }
    return true;
  }

  /**
   * 不正确的类型返回 string
   */
  static String convertType(String type) {
    if (type == null || type.trim().length == 0) {
      type = "string";
    }
    type = type.trim().toLowerCase();
    if (!isValidType(type)) {
      type = "string";
    }
    return type;
  }

  static List<String> ALL_TYPES = [
    "int",
    "double",
    "bool",
    "string"
  ];

  static bool isValidType(String type) {
    for (var s in ALL_TYPES) {
      if (s == type) {
        return true;
      }
    }
    return false;
  }

  /// 检查字符串是否有特殊字符，检查方式为 url encode 结果是否与原值相等
  ///
  /// TODO 这里需要再三确认
  static bool isValidRaw(final String s) {
    try {
      String decoded = Uri.decodeQueryComponent(s);
      if (decoded != s) {
        // 不相同了，说明原本就已经是 encode 过的了
        return true;
      }
      // 相同说明是没有 encode 过的，接着向下走
    } catch (e) {
      // 出现错误说明不是 encode 过的
    }
    try {
      // 如果 encode 之后跟之前一样，说明是正常字符
      return s == Uri.encodeQueryComponent(s);
    } catch (e) {
      return false;
    }
  }
}
