import 'package:build/src/builder/builder.dart';
import 'package:zrouter/src/compiler/compiler_data.dart';
import 'package:zrouter/src/compiler/mapper_creator.dart';

import '../../annotations.dart';
import '../../resolver/models.dart';
import 'compiler_util.dart';
import 'route_element.dart';

class RouterProcessor {
  /// import 与前缀的对应
  static final Map<String, String> importPrefix = {};

  static bool sDebug = false;

  String process(Set<RouteElement> allRoutes, BuilderOptions options) {
    StringBuffer buffer = new StringBuffer("");

    List<InnerMapper> mappers = [];
    allRoutes.forEach((r) {
      mappers.addAll(processParams(r, Logger(r, buffer)));
    });

    mappers.sort((InnerMapper m1, InnerMapper m2) {
      return m2.priority - m1.priority;
    });

    // String widgetPrefix = "${RouteElement.importPrefix}widget";

    // 写入基本的 import
    buffer.write("""
import 'package:zrouter/resolver/models.dart';
import 'package:zrouter/route_mappers.dart';
""");

    // 写入所有路由的 import
    for (var entry in CompilerData.importPrefix.entries) {
      buffer.write("import '${entry.key}' as ${entry.value};\n");
    }

    // 写入生成类
    buffer.write("""

class RouterMapperInitializer {
  static void init() {

    RouterMappers.DEFAULT_SCHEME = "${options.config["scheme"] ?? "zhihu"}";

    RouterWidgetBuilder _builder;
    Route redirect;
    Route route;
    List<Segment> rPath;
    List<Segment> path;
    Map<String, Query> rQuery;
    Map<String, Query> query;
    
    """);

    for (var value in mappers) {
      writeMapperToCode(value, buffer);
    }

    buffer.write("""
  }
}
    """);
    return buffer.toString();
  }

  Iterable<InnerMapper> processParams(RouteElement r, Logger logger) sync* {
    ScreenRoute zRoute = fromAnnotation(logger, r.annotation);
    for (var route in (zRoute.route as Set<String>).toList()) {
      InnerMapper mapper = MapperCreator.generateMapperEntity(
          logger,
          r,
          route,
          r.element,
          zRoute.paramReg,
          zRoute.priority,
          "");
      if (mapper != null) {
        yield mapper;
      } else {
        logger.error("failed");
      }
    }
  }

  void addFailedAnalyzingHeader(StringBuffer buffer, InnerMapper mapper) {
    buffer.write("\n");
    buffer.write(
        "↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑ 生成失败, url 是 ${mapper.template}, 页面是${mapper.target
            .importString + ":" +
            mapper.target.displayName} ↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑");
  }

  void writeMapperToCode(InnerMapper mapper, StringBuffer buffer) {
    buffer.writeln(
        "    // =============== mapping  ${mapper.target.importString + ":" +
            mapper.target.displayName} ===============");
    buffer.writeln("    // url is ${mapper.template}\n");

    // 生成 redirect
    if (mapper.route.redirect != null) {
      writeRouteToCode(
          buffer, mapper.route.redirect, "redirect", "rPath", "rQuery", "null");
    } else {
      buffer.writeln("redirect = null;");
    }

    // 生成 route
    writeRouteToCode(
        buffer, mapper.route, "route", "path", "query", "redirect");

    // 生成 builder，要决定是使用 constructor 还是 function 还是带参数的
    buffer.writeln(mapper.getBuilder("_builder"));

    // addMapper
    buffer.writeln(
        "RouterMappers.addMapper(Mapper(${_s(
            mapper.template)}, route, _builder, ${mapper.priority},${_s(
            mapper.module)}));\n\n");
  }

  void writeRouteToCode(StringBuffer buffer, Route route, String routeName,
      String segmentsName, String queriesName, String subRouteName) {
    List<Segment> path = route.path;
    if (path.isEmpty) {
      // path = const[]; 或者 rPath = const[];
      buffer.writeln("$segmentsName = const [];");
    } else {
      // 生成这样的代码：path = const []
      // 或者 rPath
      buffer.write("$segmentsName = const [");
      for (int i = 0; i < path.length; i++) {
        Segment segment = path[i];
        // debug(segment.toString());
        buffer.write(
            "Segment(${_s(segment.raw)}, ${_s(segment.key)}, ${_s(
                segment.value)}, ${_s(segment.type)}, ${segment.isParam}, ${_s(
                segment.reg)}),");
      }
      buffer.write("];\n");
    }

    if (route.queries.isEmpty) {
      // query = const {};
      buffer.writeln("$queriesName = const {};");
    } else {
      // query = const {
      //  'uid': Query("{uid}", "uid", "uid", null, "string", true, false, null),
      // };
      buffer.writeln("${queriesName} = const {");
      for (var entry in route.queries.entries) {
        Query query = entry.value;
        // debug(query.toString());
        //queriesName[queryKey] = Query(raw, queryKey, key, value, type, isParam, nullable, reg);
        buffer.writeln(
            "'${query.queryKey}' : Query(${_s(query.raw)}, ${_s(
                query.queryKey)}, ${_s(query.bundleKey)}, ${_s(
                query.value)}, ${_s(query.type)}, ${query.isParam}, ${query
                .nullable}, ${_s(query.reg)}),");
      }
      buffer.writeln("};");
    }
    buffer.writeln(
        "$routeName = Route(${_s(route.template)}, ${_s(route.scheme)}, ${_s(route.host)}, $segmentsName, $queriesName, $subRouteName);");
  }

  String _s(String s) {
    if (s == null) {
      return "null";
    }
    if (s.contains('"') && s.contains("'")) {
      // 此处需要固定使用单引号，且不可变更
      return "'$s'";
    }
    if (s.contains('"')) {
      return "r'$s'";
    } else {
      return 'r"$s"';
    }
  }
}
