import 'package:flutter/widgets.dart';
import 'package:test/test.dart';
import 'package:zrouter/resolver/interfaces.dart';
import 'package:zrouter/resolver/match_result.dart';
import 'package:zrouter/router.dart';

import 'demos/constructor_route.dart';
import 'demos/double_screen.dart';
import 'demos/param_reg_screen.dart';
import 'demos/priority_screen.dart';
import 'demos/redirect.dart';
import 'demos/route_mapper_generator.route.dart';
import 'demos/same_screen.dart';
import 'demos/same_screen_2.dart' as lib1;
import 'demos/segment_screen.dart';
import 'demos/simple_screen.dart';
import 'utils.dart';

void prepare() {
  RouterMapperInitializer.init();
}

void main() {
  prepare();

  test("paramReg 测试", () {
    var result = resolve(
        "zhihu://lives/122121212121212121212121?type=comments&user=john");
    expect(result, isNotNull);
    expect(result.widget.runtimeType, ParamRegScreen);

    // 不能有 admin
    result = resolve(
        "zhihu://lives/122121212121212121212121?type=comments&user=john&admin=3");
    expect(result, isNull);

    // user 不能是 peter
    result = resolve(
        "zhihu://lives/122121212121212121212121?type=comments&user=peter");
    expect(result, isNull);

    // live_id 必须是数字
    result = resolve("zhihu://lives/explore?type=comments&user=john");
    expect(result, isNull);
  });

  test("SecondPriority 优先级更高，返回 SecondPriority", () {
    var result = resolve("zhihu://priority/test");
    expect(result, isNotNull);
    expect(result.widget.runtimeType, SecondPriority);
  });

  test("DoubleRouteScreen 有两个路由注解，两个都应该解析出来", () {
    var result = resolve("zhihu://double/first?uid=3");
    expect(result.widget.runtimeType, DoubleRouteScreen);
    expect(result.args["uid"], "3");

    result = resolve("zhihu://double/second?ps=4");
    expect(result.widget.runtimeType, DoubleRouteScreen);
    expect(result.args["ps"], "4");
  });

  test("MultiRouteScreen 一个注解定义了两条路由，两个都应该解析出来", () {
    var result = resolve("zhihu://multi/first?uid=3");
    expect(result.widget.runtimeType, MultiRouteScreen);
    expect(result.args["uid"], "3");

    result = resolve("zhihu://multi/second?uid=4");
    expect(result.widget.runtimeType, MultiRouteScreen);
    expect(result.args["uid"], 4);
  });

  test("有两个都叫 SameOne 的 Widget，不应该有冲突", () {
    var result = resolve("zhihu://same/lib1/one");
    expect(result.widget.runtimeType, lib1.SameOne);

    result = resolve("zhihu://same/main/one");
    expect(result.widget.runtimeType, SameOne);
  });

  test("QueryScreen 测试", () {
    // 全测试
    var result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=2.3&boo=true&must=mmm&keys=mykey");
    expect(result.widget.runtimeType, QueryScreen);
    expect(result.args["str"], "strings");
    expect(result.args["num"], 1);
    expect(result.args["dou"], 2.3);
    expect(result.args["boo"], true);
    expect(result.args["must"], "mmm");
    expect(result.args["default"], "def");
    expect(result.args["bundleKey"], "mykey");
    expect(result.args["opt"], null);

    // 可空的也可以赋值
    result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=2.3&boo=true&must=mmm&keys=mykey&opt=3");
    expect(result.widget.runtimeType, QueryScreen);
    expect(result.args["opt"], "3");

    // default 的也可以赋值
    result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=2.3&boo=true&must=mmm&keys=mykey&default=3434");
    expect(result.widget.runtimeType, QueryScreen);
    expect(result.args["default"], "3434");

    // 不可空和必须有
    result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=2.3&boo=true&keys=mykey&opt=3");
    expect(result, isNull);

    // 任一类型不对都不行
    // num 不是 Int
    result = resolve(
        "zhihu://query/king?str=strings&num=e&dou=2.3&boo=true&must=mmm&keys=mykey");
    expect(result, isNull);
    // dou 不是 double
    result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=k&boo=true&must=mmm&keys=mykey");
    expect(result, isNull);
    // boo 不是 bool
    result = resolve(
        "zhihu://query/king?str=strings&num=1&dou=2.3&boo=haha&must=mmm&keys=mykey");
    expect(result, isNull);
  });

  test("Path 测试", () {
    // 全测试
    var result = resolve("zhihu://path/strings/1/2.3/true/path/screen?a=b");
    expect(result.widget.runtimeType, PathScreen);
    expect(result.args["str"], "strings");
    expect(result.args["num"], 1);
    expect(result.args["dob"], 2.3);
    expect(result.args["boo"], true);
    expect(result.args["a"], "b");

    // 任一类型不对都不行
    // num 不是 Int
    // result = resolve("zhihu://path/strings/1/2.3/true/path/screen?a=b");
    result = resolve("zhihu://path/strings/e/2.3/true/path/screen?a=b");
    expect(result, isNull);
    // dou 不是 double
    result = resolve("zhihu://path/strings/1/e/true/path/screen?a=b");
    expect(result, isNull);
    // boo 不是 bool
    result = resolve("zhihu://path/strings/1/2.3/haha/path/screen?a=b");
    expect(result, isNull);
  });

  test("方法测试", () {
    var result = resolve("zhihu://cons/method/333?str=myname");
    expect(result.widget.runtimeType, Text);
    expect((result.widget as Text).data, "myname_333");
    expect(result.args["i"], 333);
    expect(result.args["str"], "myname");
  });

  test("自定义 Constructor 测试，走到 ConstructorScreen.old", () {
    var result = resolve(
        "zhihu://cons/params?intValue=3&stringValue=4343&doubleValue=3.3&boolValue=true");
    expect(result.widget.runtimeType, ConstructorScreen);
    expect(result.args["intValue"], 3);
    expect(result.args["stringValue"], "4343");
    expect(result.args["doubleValue"], 3.3);
    expect(result.args["boolValue"], true);
    expect((result.widget as ConstructorScreen).intValue, 3);
    expect((result.widget as ConstructorScreen).stringValue, "4343");
    expect((result.widget as ConstructorScreen).doubleValue, 3.3);
    expect((result.widget as ConstructorScreen).boolValue, true);
    expect((result.widget as ConstructorScreen).fromOld, "old");
  });

  test("自定义 Constructor 测试", () {
    var result = resolve(
        "zhihu://cons/params2?intValue=3&stringValue=4343&doubleValue=3.3&boolValue=false");
    expect(result.widget.runtimeType, Constructor2Screen);

    expect(result.args["intValue"], 3);
    expect(result.args["stringValue"], "4343");
    expect(result.args["doubleValue"], 3.3);
    expect(result.args["boolValue"], false);

    expect((result.widget as Constructor2Screen).intValue, 3);
    expect((result.widget as Constructor2Screen).stringValue, "4343");
    expect((result.widget as Constructor2Screen).doubleValue, 3.3);
    expect((result.widget as Constructor2Screen).boolValue, false);
    expect((result.widget as Constructor2Screen).fromCons, "Yes");
  });

  test("redirect 测试", () {
    var result = resolve(
        "zhihu://answer/answer_idddd?uid=uidddd&sample_id=sample_idddd");
    expect(result.widget.runtimeType, RedirectScreen);
    expect(result.args["answer_id"], "answer_idddd");
    expect(result.args["uid"], "uidddd");
    expect(result.args["sample_id"], "sample_idddd");

    // 重定向后，url 也变了
    expect(result.args[Router.KEY_ROUTER_RAW_URL],
        'https://zhihu.com/uidddd/answer/answer_idddd?sample_id=sample_idddd');
    expect(result.args[Router.KEY_ROUTER_REDIRECT_FROM],
        "zhihu://answer/answer_idddd?uid=uidddd&sample_id=sample_idddd");

    result = resolve(
        "zhihu://answer2/answer_idddd?uid=uidddd&sample_id=sample_idddd");
    expect(result.widget.runtimeType, RedirectNamedScreen);
    expect((result.widget as RedirectNamedScreen).answer_id, "answer_idddd");
    expect((result.widget as RedirectNamedScreen).uid, "uidddd");
    expect((result.widget as RedirectNamedScreen).sample_id, "sample_idddd");
  });

  test("超简单测试", () {
    var result = resolve("simple/scene");
    expect(result.widget.runtimeType, SimpleScreenOne);
    result = resolve("simple");
    expect(result.widget.runtimeType, SimpleScreenTwo);
  });

  test("default scheme 测试", () {
    var result = resolve("zhihu://simple/scene");
    expect(result.widget.runtimeType, SimpleScreenOne);
    result = resolve("simple");
    expect(result.widget.runtimeType, SimpleScreenTwo);

    result = resolve("same/main/one");
    expect(result.widget.runtimeType, SameOne);
  });

  test("equal checker 测试", () {
    Router.setChecker(Zhihu2Checker());
    var result = resolve("zhihu2://simple/scene");
    expect(result.widget.runtimeType, SimpleScreenOne);
  });

//  test("超简单测试", () {
//    Router.addFilter(AlwaysFail());
//    var result = resolve("simple/scene");
//    expect(result, isNull);
//  });
}

class AlwaysFail extends Filter {
  @override
  bool test(MatchResult result) {
    return false;
  }
}

class Zhihu2Checker extends EqualChecker {
  @override
  bool checkEquals(
      String scheme, String templateScheme, String host, String templateHost) {
    if (host != templateHost) {
      return false;
    }
    return scheme == templateScheme ||
        (templateScheme == "zhihu" && scheme == "zhihu2");
  }

  @override
  bool intercept(
      String scheme, String templateScheme, String host, String templateHost) {
    return true;
  }
}
